import typing

import numpy
from pandas import DataFrame

from roguewave import WaveSpectrum1D, WaveSpectrum2D
from roguewave.wavespectra import spec2d_from_spec1d
from roguewave.wavespectra.operators import spectrum1D_time_filter, \
    spectrum2D_time_filter
from roguewave.wavespectra.partitioning import Partition, \
    partition_spectrum, merge_partitions, find_index_closest_partition
from roguewave.wavespectra.spectrum2D import empty_spectrum2D_like
from roguewave.wavespectra.wavespectrum import BulkVariables, WaveSpectrum


class SeaSwellData(typing.TypedDict):
    """

    """
    sea: DataFrame
    total_swell: DataFrame
    partitions: typing.List[DataFrame]


class BulkPartitionVariables():
    def __init__(self, partitions: typing.List[Partition]):
        # total_swell_spectrum = empty_spectrum2D_like(partitions[0].spectrum)
        self._raw_partitions = partitions

    @property
    def sea(self) -> BulkVariables:
        for x in self._raw_partitions:
            if x.is_sea_partition():
                return x.spectrum.bulk_variables()
        else:
            bulk = BulkVariables(None)
            bulk.timestamp = self._raw_partitions[0].spectrum.timestamp
            bulk.latitude = self._raw_partitions[0].spectrum.latitude
            bulk.longitude = self._raw_partitions[0].spectrum.longitude
            return bulk

    @property
    def total_swell(self) -> BulkVariables:
        total_swell_spectrum = empty_spectrum2D_like(
            self._raw_partitions[0].spectrum)

        ii = 0
        for x in self._raw_partitions:
            if not x:
                continue

            if x.is_swell_partition():
                ii += 1
                total_swell_spectrum.variance_density = total_swell_spectrum.variance_density + x.spectrum.variance_density

        if ii > 0:
            total_swell_spectrum._update()
            return total_swell_spectrum.bulk_variables()
        else:
            bulk = BulkVariables(None)
            bulk.timestamp = self._raw_partitions[0].spectrum.timestamp
            bulk.latitude = self._raw_partitions[0].spectrum.latitude
            bulk.longitude = self._raw_partitions[0].spectrum.longitude
            return bulk

    @property
    def partitions(self) -> typing.List[BulkVariables]:
        out = []

        for x in self._raw_partitions:
            if x:
                bulk = x.spectrum.bulk_variables()
            else:
                bulk = BulkVariables(None)
                bulk.timestamp = self.total_swell.timestamp
                bulk.latitude = self.total_swell.latitude
                bulk.longitude = self.total_swell.longitude
            out.append(bulk)

        return out

    @property
    def number_of_partitions(self):
        return len(self._raw_partitions)

    def renumber(self, mapping: typing.List[int]):

        assert len(mapping) == self.number_of_partitions
        new = []
        for new_index, old_index in enumerate(mapping):
            new.append(self._raw_partitions[old_index])
        self._partitions_swell = new

    def match_exists(self, partition:Partition):
        return self.find_matching_partition(partition) >= 0

    def find_matching_partition(self, partition:Partition,pr=False):

        correlation = numpy.zeros(len(self._raw_partitions))

        source = partition.spectrum.variance_density[:] - numpy.mean(partition.spectrum.variance_density[:])
        source =source.flatten()
        for ii, part in enumerate(self._raw_partitions):
            if part is not None:
                target = part.spectrum.variance_density[:] - numpy.mean(part.spectrum.variance_density[:])
                target=target.flatten()
                correlation[ii] = numpy.corrcoef( source.flatten(), target.flatten(), )[0,1]
            else:
                correlation[ii] = 0

        i = numpy.nanargmax(correlation)

        if pr:
            #print(correlation)
            pass
        if correlation[i] >= 0.7:
            return i
        else:
            return -1

    def fill_to(self, n):
        number_of_partitions = len(self._raw_partitions)
        if number_of_partitions == n:
            return

        if number_of_partitions > n:
            self._raw_partitions = self._raw_partitions[:n]
            return

        for ii in range(number_of_partitions, n):
            self._raw_partitions.append(None)

    def closest_index(self,source_index):
        target_index, _ = find_index_closest_partition(source_index,
                                                    self._raw_partitions)
        return target_index

    def merge(self, source_index , target_index):
        merge_partitions(source_index, target_index, self._raw_partitions)
        return


def _gen_dataframe_from_partitions(
        bulk: typing.List[BulkVariables]) -> DataFrame:
    """
    From a partition, create a dataframe.
    :param bulk:
    :return:
    """
    n = len(bulk)
    data = {}
    properties = ['m0', 'hm0', 'tm01', 'tm02', 'peak_period', 'peak_direction',
                  'peak_spread',
                  'bulk_direction', 'bulk_spread', 'peak_frequency',
                  'peak_wavenumber', 'latitude', 'longitude', 'timestamp']

    for key in properties:
        if key == 'timestamp':
            data[key] = numpy.empty((n,), dtype=object)
        else:
            data[key] = numpy.zeros((n,))

    for i, b in enumerate(bulk):
        for key in properties:
            data[key][i] = getattr(b, key)

    return DataFrame.from_dict(data)


def _homogonize_bulk_swell(bulk: typing.List[BulkPartitionVariables]) -> \
        typing.List[BulkPartitionVariables]:
    """
    Ensure that all entries in the list have the same number of swell partitions
    :param bulk:
    :return:
    """
    max_partitions = numpy.array([b.number_of_partitions for b in bulk]).max()
    for b in bulk:
        b.fill_to(max_partitions)
    return bulk


def match_partitions(bulk_partitions: typing.List[BulkPartitionVariables]) -> \
typing.List[BulkPartitionVariables]:
    for index, partition in enumerate(bulk_partitions):
        if index == 0:
            continue

        mapping = numpy.zeros((partition.number_of_partitions,),
                              dtype='int32') - 1
        previous = bulk_partitions[index - 1]

        mapped = []
        notmapped = []
        empty = []
        for old_index, swell in enumerate(partition.partitions):
            if numpy.isnan(swell.hm0):
                empty.append(old_index)
            else:
                index = previous.find_matching_partition(
                    partition._raw_partitions[old_index])
                if index >= 0:
                    mapped.append((old_index, index))
                else:
                    notmapped.append(old_index)

        ii = 0
        for map in mapped:
            mapping[map[1]] = map[0]

        jj = -1
        if notmapped:
            for index, map in enumerate(mapping):
                if map < 0:
                    jj += 1
                    mapping[index] = notmapped[jj]
                    if jj == len(notmapped)-1:
                        break

        jj = -1
        if empty:
            for index, map in enumerate(mapping):
                if map < 0:
                    jj += 1
                    mapping[index] = empty[jj]
        partition.renumber(mapping)

    for index, bulk in enumerate(bulk_partitions):

        iu = index + 1
        id = index - 1
        if index == len(bulk_partitions) - 1:
            iu = id

        if index == 0:
            id = iu

        prev = bulk_partitions[id]
        next = bulk_partitions[iu]
        #

        to_merge = {}
        if bulk.number_of_partitions <= 1:
            continue

        for ipart, partition in enumerate(bulk.partitions):
            #
            if numpy.isnan(partition.hm0):
                continue

            match_in_prev = prev.match_exists(bulk._raw_partitions[ipart])
            match_in_next = next.match_exists(bulk._raw_partitions[ipart])

            if index == 3:
                print('---')
                print(index,ipart, match_in_prev, match_in_next)
                print(index,ipart, prev.find_matching_partition(bulk._raw_partitions[ipart],pr=True) )
                print(index, ipart, next.find_matching_partition(
                    bulk._raw_partitions[ipart],pr=True))


            if not (  match_in_prev or match_in_next ):
                target = bulk.closest_index(ipart)
                if target in to_merge:
                    target = to_merge[target]

                if not index==target:
                    to_merge[ipart] = target

        for source,target in to_merge.items():
            bulk.merge(source,target)

        to_merge = [ x for x in to_merge]
        for index in sorted(to_merge, reverse=True):
            del bulk._raw_partitions[index]

    return bulk_partitions


def sea_swell_data(spectra: typing.List[WaveSpectrum],
                   time_filter=True, verbose=True,
                   filter_window=None) -> SeaSwellData:
    """

    :param spectra: list of wavespectra objects.
    :return: dictionary (typed: SeaSwellData) that contains:
        { "sea": pandas dataframe
          "total_swell": pandas dataframe
          "partitions_swell": list[pandas dataframe]

    """
    bulk = []
    # For each partition do:

    if time_filter:
        if isinstance(spectra[0], WaveSpectrum1D):
            spectra: typing.List[WaveSpectrum1D]
            spectra = spectrum1D_time_filter(spectra, filter_window)
        elif isinstance(spectra[0], WaveSpectrum2D):
            spectra: typing.List[WaveSpectrum2D]
            spectra = spectrum2D_time_filter(spectra, filter_window)

    for ii, spectrum in enumerate(spectra):

        if verbose:
            print(ii)

        if isinstance(spectrum, WaveSpectrum1D):
            spec2d = spec2d_from_spec1d(spectrum)
        elif isinstance(spectrum, WaveSpectrum2D):
            spec2d = spectrum
        else:
            raise Exception('Unknown spectral type')

        # Partition the spectrum
        partition, _ = partition_spectrum(spec2d)

        # calculate bulk variables for each partition
        bulk.append(BulkPartitionVariables(partition))

    bulk = _homogonize_bulk_swell(bulk)
    bulk = match_partitions(bulk)

    sea = _gen_dataframe_from_partitions([b.sea for b in bulk])
    total_swell = _gen_dataframe_from_partitions([b.total_swell for b in bulk])

    #
    # Create a list of bulk descriptions of the swell, and make sure that all
    # bulk descriptions contain the same number of swells. This makes processing
    # easier (no other reason).
    #
    bulk = _homogonize_bulk_swell(bulk)

    # now all data have the same number of swell partitions
    max_partitions = bulk[0].number_of_partitions

    # For each partition, create the dataframe.
    partitions_swell = []
    for ii in range(0, max_partitions):
        partitions_swell.append(
            _gen_dataframe_from_partitions(
                [b.partitions[ii] for b in bulk])
        )

    # Return Data
    return SeaSwellData(sea=sea, total_swell=total_swell,
                        partitions=partitions_swell)



from .wavefields import wave_fields_from, filter_fields, create_graph, \
    bulk_parameters_partitions
from .partitioning import partition_spectrum, default_partition_config
from roguewave.wavespectra import spec1d_from_spec2d, WaveSpectrum1D
from roguewave.wavespectra.operators import spectrum1D_time_filter
from roguewave.wavespectra.estimators import spec2d_from_spec1d
from typing import List, Dict, Union
from datetime import timedelta
from pandas import DataFrame
from roguewave.wavespectra import WaveSpectrum2D
from typing import overload

default_config = {
    'smoothInTime': False,
    'estimator': {
        'method': 'mem2',
        'numberOfDirections': 36,
        'frequencySmoothing': False,
        'smoothingLengthscale': 1
    },
    'partitionConfig': default_partition_config,
    'fieldFiltersettings': {
        'filter': True,
        'maxDeltaPeriod': 2,
        'maxDeltaDirection': 20
    }
}


# -----------------------------------------------------------------------------
#                       Boilerplate Interfaces
# -----------------------------------------------------------------------------
@overload
def get_bulk_partitions_from_spectral_partitions(
        spectral_partitions: Dict[str, List[List[WaveSpectrum2D]]]
) -> Dict[str, List[DataFrame]]: ...

@overload
def get_bulk_partitions_from_spectral_partitions(
        spectral_partitions: List[List[WaveSpectrum2D]]
) -> List[DataFrame]: ...

@overload
def get_spectral_partitions_from_observations(
        spectra: Dict[str, List[WaveSpectrum1D]],
        minimum_duration: timedelta,
        config=None,
        verbose=False) -> Dict[str, List[List[WaveSpectrum2D]]]: ...


@overload
def get_spectral_partitions_from_observations(
        spectra: List[WaveSpectrum1D],
        minimum_duration: timedelta,
        config=None,
        verbose=False) -> List[List[WaveSpectrum2D]]: ...


@overload
def get_bulk_partitions_from_observations(
        spectra: Dict[str, List[WaveSpectrum1D]],
        minimum_duration: timedelta,
        config=None,
        verbose=False) -> Dict[str, List[DataFrame]]: ...


@overload
def get_bulk_partitions_from_observations(
        spectra: List[WaveSpectrum1D],
        minimum_duration: timedelta,
        config=None,
        verbose=False) -> List[DataFrame]: ...


# -----------------------------------------------------------------------------
#                       Implementationb
# -----------------------------------------------------------------------------
def get_bulk_partitions_from_spectral_partitions(
        spectral_partitions: Union[Dict[str, List[List[WaveSpectrum2D]]], List[
            List[WaveSpectrum2D]]]) -> Union[Dict[
                                                 str, List[DataFrame]], List[
                                                 DataFrame]]:
    if isinstance(spectral_partitions, dict):
        output = {}
        for key in spectral_partitions:
            output[key] = bulk_parameters_partitions(spectral_partitions[key])

    elif isinstance(spectral_partitions, list):
        output = bulk_parameters_partitions(spectral_partitions)
    else:
        raise Exception('Cannot process input')

    return output


def get_bulk_partitions_from_observations(
        spectra: Union[Dict[str, List[WaveSpectrum1D]], List[WaveSpectrum1D]],
        minimum_duration: timedelta,
        config=None,
        verbose=False) -> Union[Dict[str, List[DataFrame]], List[DataFrame]]:
    #
    spectral_partitions = get_spectral_partitions_from_observations(
        spectra, minimum_duration, config, verbose)

    if isinstance(spectra, dict):
        output = {}
        for key in spectral_partitions:
            output[key] = bulk_parameters_partitions(spectral_partitions[key])

    elif isinstance(spectra, list):
        output = bulk_parameters_partitions(spectral_partitions)
    else:
        raise Exception('Cannot process input')

    return output


def get_spectral_partitions_from_observations(
        spectra: Union[Dict[str, List[WaveSpectrum1D]], List[WaveSpectrum1D]],
        minimum_duration: timedelta,
        config=None,
        verbose=False) -> Union[
    Dict[str, List[List[WaveSpectrum2D]]], List[List[WaveSpectrum2D]]]:
    #

    if isinstance(spectra, dict):
        output = {}
        for key, item in spectra.items():
            output[key] = partition_observations_spectra(item,
                                                         minimum_duration,
                                                         config, verbose)
    elif isinstance(spectra, list):
        output = partition_observations_spectra(spectra,
                                                minimum_duration,
                                                config, verbose)
    else:
        raise Exception('Cannot process input')

    return output


def _print(verbose, *narg, **kwargs):
    if verbose:
        print(*narg, **kwargs)


def partition_observations_bulk(spectra: List[WaveSpectrum1D],
                                minimum_duration: timedelta,
                                config=None, verbose=False) -> List[DataFrame]:
    wave_fields = partition_observations_spectra(spectra, minimum_duration,
                                                 config=config,
                                                 verbose=verbose)
    return bulk_parameters_partitions(wave_fields)


def partition_observations_spectra(spectra: List[WaveSpectrum1D],
                                   minimum_duration: timedelta,
                                   config=None, verbose=False) -> List[
    List[WaveSpectrum2D]]:
    if config:
        for key in config:
            assert key in default_config, f"{key} is not a valid conficuration entry"

        config = default_config | config
    else:
        config = default_config

    # Step 1: Pre-Processing
    # :
    # Prior to constructing spectra - smoothing the wave field can help create
    # more stable results.

    _print(verbose, '*** Partitioning Data ***\n' + 80 * '-' + '\n')
    if config['smoothInTime']:
        _print(verbose, ' - Smoothing in time')
        spectra = spectrum1D_time_filter(spectra)

    # Step 2: Create 2D wavefields from 1D spectra using a spectral estimator
    spectra2D = []
    _print(verbose, ' - Create 2D Spectra')
    for index, spectrum in enumerate(spectra):
        _print(verbose, f'\t {index:05d} out of {len(spectra)}')
        spectra2D.append(
            spec2d_from_spec1d(
                spectrum,
                method=config['estimator']['method'],
                number_of_directions=config['estimator']['numberOfDirections'],
                frequency_smoothing=config['estimator']['frequencySmoothing'],
                smoothing_lengthscale=config['estimator'][
                    'smoothingLengthscale']
            )
        )

    # Step 3: Partition the data
    _print(verbose, ' - Partition Data')
    raw_partitions = []
    for index, spectrum in enumerate(spectra2D):
        _print(verbose, f'\t {index:05d} out of {len(spectra2D)}')
        partitions, _ = partition_spectrum(spectrum, config['partitionConfig'])
        raw_partitions.append(partitions)

    # Step 4: create a graph
    _print(verbose, ' - Create Graph')
    graph = create_graph(raw_partitions, minimum_duration)

    # Step 5: create wave field from the graph
    _print(verbose, ' - Create Wave Fields From Graph')
    wave_fields = wave_fields_from(graph)

    # Step 6: Postprocessing

    # Apply a filter on the bulk parameters
    _print(verbose, ' - Apply Bulk Filter')
    if config['fieldFiltersettings']['filter']:
        wave_fields = filter_fields(
            wave_fields,
            min_duration=minimum_duration,
            max_delta_period=config['fieldFiltersettings']['maxDeltaPeriod'],
            max_delta_direction=config['fieldFiltersettings'][
                'maxDeltaDirection']
        )
    _print(verbose, '*** Done ***\n' + 80 * '-' + '\n\n')
    return wave_fields



from .spectrum2D import WaveSpectrum2D
from .wavespectrum import WaveSpectrum
from .estimators import spec1d_from_spec2d
from .parametric import pierson_moskowitz_frequency


def is_sea_spectrum(spectrum: WaveSpectrum) -> bool:
    """
    Identify whether or not it is a sea partion. Use 1D method for 2D
    spectra in section 3 of:

    Portilla, J., Ocampo-Torres, F. J., & Monbaliu, J. (2009).
    Spectral partitioning and identification of wind sea and swell.
    Journal of atmospheric and oceanic technology, 26(1), 107-122.

    :return: boolean indicting it is sea
    """

    if isinstance(spectrum, WaveSpectrum2D):
        spectrum = spec1d_from_spec2d(spectrum)

    peak_index = spectrum.peak_index()
    peak_frequency = spectrum.frequency[peak_index]

    return pierson_moskowitz_frequency(
        peak_frequency, peak_frequency) <= spectrum.variance_density[
               peak_index]


def is_swell_spectrum(spectrum: WaveSpectrum) -> bool:
    """
    Identify whether or not it is a sea partion. Use 1D method for 2D
    spectra in section 3 of:

    Portilla, J., Ocampo-Torres, F. J., & Monbaliu, J. (2009).
    Spectral partitioning and identification of wind sea and swell.
    Journal of atmospheric and oceanic technology, 26(1), 107-122.

    :return: boolean indicting it is sea
    """

    return not is_sea_spectrum(spectrum)





"""
This file is part of pysofar: A client for interfacing with Sofar Oceans Spotter API

Contents: 2D Spectrum Class

Copyright (C) 2022
Sofar Ocean Technologies

Authors: Pieter Bart Smit
"""
import numpy
import numpy.ma
from .wavespectrum import WaveSpectrum, WaveSpectrumInput
from roguewave.wavetheory.lineardispersion import inverse_intrinsic_dispersion_relation
from typing import List, Union
from roguewave.tools import datetime_to_iso_time_string
import scipy.ndimage
from numpy.ma import MaskedArray
import typing


class WaveSpectrum2DInput(WaveSpectrumInput):
    directions: Union[List[float], numpy.ndarray]

class WaveSpectrum2D(WaveSpectrum):
    def __init__(self,
                 wave_spectrum2D_input:WaveSpectrum2DInput
                 ):

        super().__init__(wave_spectrum2D_input)
        self.direction = numpy.array(wave_spectrum2D_input['directions'])
        self._frequency_peak_indices = None
        self._direction_peak_indices = None
        self._update()

    @WaveSpectrum.variance_density.setter
    def variance_density(self,val:numpy.ndarray):
        mask = val < 0
        self._variance_density = MaskedArray(val,mask=mask)
        self._update()

    def _update(self):
        if self.direction is not None:
            self._a1 = self._directional_moment('a', 1, normalized=True)
            self._b1 = self._directional_moment('b', 1, normalized=True)
            self._a2 = self._directional_moment('a', 2, normalized=True)
            self._b2 = self._directional_moment('b', 2, normalized=True)
            self._e = self._directional_moment('zero', 0, normalized=False)
            self._frequency_peak_indices, self._direction_peak_indices = self.find_peak_indices(update=True)

    def _delta(self):
        angles = self.direction
        forward_diff = (numpy.diff(angles, append=angles[0]) + 180) % 360 - 180
        backward_diff = (numpy.diff(angles,
                                    prepend=angles[-1]) + 180) % 360 - 180
        return (forward_diff + backward_diff) / 2

    def _directional_moment(self, kind='zero', order=0,
                            normalized=True) -> numpy.array:
        delta = self._delta()
        if kind == 'a':
            harmonic = numpy.cos(self.radian_direction * order) * delta
        elif kind == 'b':
            harmonic = numpy.sin(self.radian_direction * order) * delta
        elif kind == 'zero':
            harmonic = delta
        else:
            raise Exception('Unknown moment')
        density = self.variance_density.filled(0)

        values = numpy.sum(density * harmonic[None, :], axis=-1)

        if normalized:
            scale = numpy.sum(density * delta[None, :], axis=-1)
            scale[scale == 0] = 1
        else:
            scale = 1
        return values / scale

    def frequency_moment(self, power: int, fmin=0, fmax=numpy.inf) -> float:
        range = self._range(fmin, fmax)
        return numpy.trapz(self.e[range] * self.frequency[range] ** power,
                           self.frequency[range])

    def _create_wave_spectrum_input(self)->WaveSpectrum2DInput:
        data = self.variance_density.filled(-1)
        return WaveSpectrum2DInput(
            frequency=list(self.frequency),
            directions=list(self.direction),
            varianceDensity=data.tolist(),
            timestamp=datetime_to_iso_time_string(self.timestamp),
            latitude=self.latitude,
            longitude=self.longitude,
        )

    def copy(self)->"WaveSpectrum2D":
        input = WaveSpectrum2DInput(
            frequency=self.frequency.copy(),
            directions=self.direction.copy(),
            varianceDensity=self.variance_density.copy(),
            timestamp=self.timestamp,
            latitude=self.latitude,
            longitude=self.longitude
        )
        return WaveSpectrum2D(input)

    def __add__(self, other:"WaveSpectrum2D")->"WaveSpectrum2D":
        spectrum = self.copy()

        self_dens = numpy.ma.MaskedArray(spectrum.variance_density)
        other_dens = numpy.ma.array(other.variance_density)

        mask = self_dens.mask & other_dens.mask

        spectrum.variance_density = numpy.ma.MaskedArray(self_dens.filled(0) +
            other_dens.filled(0), mask)
        return spectrum

    def __sub__(self, other: "WaveSpectrum2D") -> "WaveSpectrum2D":
        spectrum = self.copy()
        self_dens = numpy.ma.MaskedArray(spectrum.variance_density)
        other_dens = numpy.ma.array(other.variance_density)

        mask = self_dens.mask & other_dens.mask

        spectrum.variance_density = numpy.ma.MaskedArray(self_dens.filled(0) -
            other_dens.filled(0), mask)
        return spectrum

    def __neg__(self, other: "WaveSpectrum2D") -> "WaveSpectrum2D":
        spectrum = self.copy()
        spectrum.variance_density = -spectrum.variance_density
        return spectrum

    def extract(self,mask:numpy.ndarray, minimum_ratio=None)->"WaveSpectrum2D":
        spectrum = self.copy()

        if minimum_ratio:
            ii,jj = self.find_peak_indices()
            peak_density = self._variance_density[ ii[0], jj[0] ]
            if peak_density > 0.0:
                mask = mask & (spectrum.variance_density/peak_density > minimum_ratio)
            else:
                print('eh')
        density = numpy.ma.masked_array( spectrum.variance_density, mask=~mask)
        spectrum.variance_density = density
        return spectrum

    def peak_direction(self):
        _, index = self.find_peak_indices()
        return self.direction[index[0]]

    def peak_wavenumber(self, depth=numpy.inf):
        index, _ = self.find_peak_indices()

        return inverse_intrinsic_dispersion_relation(
            self.radian_frequency[index[0]], depth)

    def peak_wavenumber_east(self, depth=numpy.inf):
        wave_number    = self.peak_wavenumber()
        wave_direction = self.peak_direction() * numpy.pi/180
        return wave_number * numpy.cos(wave_direction)

    def peak_wavenumber_north(self, depth=numpy.inf):
        wave_number    = self.peak_wavenumber()
        wave_direction = self.peak_direction() * numpy.pi/180
        return wave_number * numpy.cos(wave_direction)

    def find_peak_indices(self,update=False):
        """
        Find the peaks of the frequency-direction spectrum.
        :param density: 2D variance density
        :return: List of indices indicating the maximum
        """
        if not update and self._frequency_peak_indices is not None:
            return self._frequency_peak_indices, self._direction_peak_indices

        density = self.variance_density

        # create a search region
        neighborhood = scipy.ndimage.generate_binary_structure(density.ndim, 2)

        # Set the local neighbourhood to the the maximum value, use wrap.
        # Note that technically frequency wrapping is incorrect- but unlikely to
        # cause issues. First we apply the maximum filter .with 9 point footprint
        # to set each pixel of the output to the local maximum
        filtered = scipy.ndimage.maximum_filter(
            density, footprint=neighborhood,mode='wrap'
        )

        # Then we find maxima based on equality with input array
        maximum_mask = filtered == density

        # Remove possible background (0's)
        background_mask = density == 0

        # Remove peaks in background
        maximum_mask[background_mask] = False

        # return indices of maxima
        ii, jj = numpy.where(maximum_mask)

        # sort from lowest maximum value to largest maximum value.
        sorted = numpy.flip(numpy.argsort(density[ii, jj]))
        ii = ii[sorted]
        jj = jj[sorted]
        if len(ii) == 0:
            ii = numpy.array([0])
            jj = numpy.array([0])
        self._frequency_peak_indices = ii
        self._direction_peak_indices = jj
        return ii,jj


def empty_spectrum2D_like(spectrum:WaveSpectrum2D)->WaveSpectrum2D:
    input = WaveSpectrum2DInput(
        frequency=spectrum.frequency.copy(),
        varianceDensity=numpy.zeros_like(spectrum.variance_density),
        timestamp=spectrum.timestamp,
        latitude=spectrum.latitude,
        longitude=spectrum.longitude,
        directions=spectrum.direction.copy()
    )
    return WaveSpectrum2D(input)







from .externaldata import sofarspectralapi
from .externaldata.spotter import get_spectrum_from_sofar_spotter_api
from .wavespectra.spectrum1D import WaveSpectrum1D, WaveSpectrum1DInput
from .wavespectra.spectrum2D import WaveSpectrum2D, WaveSpectrum2DInput
from .wavespectra import spectrum1D, spectrum2D
from .wavespectra.estimators import convert_to_1d_spectrum,convert_to_2d_spectrum
from .wavespectra.io import load_spectrum, save_spectrum
from .wavespectra.partitioning.observations import \
    partition_observations_spectra, partition_observations_bulk
from .wavespectra.io import load_spectrum, save_spectrum
from .wavespectra.partitioning.observations import \
    get_bulk_partitions_from_observations, \
    get_spectral_partitions_from_observations, \
    get_bulk_partitions_from_spectral_partitions



"""
Contents: Partitioning

Copyright (C) 2022
Sofar Ocean Technologies

Authors: Pieter Bart Smit
======================

Partitioning routines that can be used to partition the spectrum.

Classes:

- `SeaSwellData`, return type that contains the partitioned data.
- `Partition`, data class that describes a partition

Functions:

- `neighbours`,

How To Use This Module
======================
(See the individual functions for details.)

1. Import it: ``import partitioning`` or ``from partitioning import ...``.
2.
"""

from roguewave.wavespectra.spectrum2D import WaveSpectrum2D, \
    empty_spectrum2D_like
import numpy
import typing
import numba
from scipy.ndimage import maximum_filter, generate_binary_structure

default_partition_config = {
    'minimumEnergyFraction': 0.01,
    'minimumDensityRatio': 0.
}


@numba.njit(cache=True)
def neighbours(peak_direction_index: int, peak_frequency_index: int,
               number_of_directions: int, number_of_frequencies: int,
               diagonals=True) -> typing.Tuple[numpy.ndarray, numpy.ndarray]:
    """
    Find all indices that correspond to the 2D neighbours of a given point
    with indices (i, j) in a rectangular raster spanned by frequencies and
    directions. There are potentially 8 neighbours:

            + (i+1,j-1)   + (i+1,   j)   +  (i+1,j+1)    ( + = neighbour, x is point )

            + (i  ,j-1)   x (i  ,   j)   +  (i  ,j+1)

            + (i-1,j-1)   + (i-1,   j)   +  (i-1,j+1)

    resulting in output:

            frequency_indices = [i  , i  , i-1, i-1, i-1, i+1, i+1, i+1]
            direction_indices = [j-1, j+1, j-1, j  , j+1, j-1, j  , j+1]

    Note that due to directional wrapping j-1 and j+1 always exist; but
    i+1 and i-1 may not. The ordering of the output is due to  ease of
    implementation

    :param peak_direction_index: central direction index
    :param peak_frequency_index: central frequency index
    :param number_of_directions: number of direction in the grid
    :param number_of_frequencies: number of frequencies in the grid
    :param diagonals: (optional), whether or not to include the diagonal indices
    as neighbours.
    :return: ( frequency_indices, direction_indices  )
    """

    # First we need to calculate the next/prev angles in frequency ann
    # direction space, noting that direction space is periodic.
    next_direction_index = (peak_direction_index + 1) % number_of_directions
    prev_direction_index = (peak_direction_index - 1) % number_of_directions
    prev_frequency_index = peak_frequency_index - 1
    next_frequency_index = peak_frequency_index + 1

    # Add the
    frequency_indices = [peak_frequency_index, peak_frequency_index]
    direction_indices = [prev_direction_index, next_direction_index]

    if diagonals:
        # The diagonals are included
        if peak_frequency_index > 0:
            # add the neighbours at the previous frequency
            frequency_indices += [prev_frequency_index, prev_frequency_index,
                                  prev_frequency_index]
            direction_indices += [prev_direction_index, peak_direction_index,
                                  next_direction_index]

        if peak_frequency_index < number_of_frequencies - 1:
            # add the neighbours at the next frequency
            frequency_indices += [next_frequency_index, next_frequency_index,
                                  next_frequency_index]
            direction_indices += [prev_direction_index, peak_direction_index,
                                  next_direction_index]
    else:
        # The diagonals are not included

        if peak_frequency_index > 0:
            # add the neighbours at the previous frequency
            frequency_indices += [prev_frequency_index]
            direction_indices += [peak_direction_index]
        if peak_frequency_index < number_of_frequencies - 1:
            # add the neighbours at the next frequency
            frequency_indices += [next_frequency_index]
            direction_indices += [peak_direction_index]

    return numpy.array(frequency_indices, dtype='int64'), numpy.array(
        direction_indices, dtype='int64')


NOT_ASSIGNED = -1


@numba.njit(cache=True)
def floodfill(frequency: numpy.ndarray, direction: numpy.ndarray,
              spectral_density: numpy.ndarray, min_val=0.0) -> typing.Tuple[
    numpy.ndarray, typing.Dict[int, typing.List[int]]]:
    """
    Flood fill algorithm. We try to find the region that belongs to a peak according
    to inverse watershed.

    :param spectral_density: 2d-ndarray, first dimension frequency, second direction
    :param partition_label: 2d-integer-ndarray, of same shape as [spectral_density].
    for each entry contains label to which the current entry belongs. Negative
    label is unasigned.

    :param peak_frequency_index: frequency index of local peak
    :param peak_direction_index: direction index of local peak
    :return:

    We start at a local peak indicated by [peak_frequency_index]
    and [peak_direction_index], label that peak with the given [label]


    """

    number_of_directions = spectral_density.shape[1]
    number_of_frequencies = spectral_density.shape[0]

    proximate_partitions = {}  # type: typing.Dict[int,typing.List[int]]
    # proximate_partitions = []

    partition_label = numpy.zeros(
        (number_of_frequencies, number_of_directions),
        dtype='int32') + NOT_ASSIGNED

    current_label = 0
    assigned = False
    for start_frequency_index in range(0, number_of_frequencies):
        for start_direction_index in range(0, number_of_directions):
            if spectral_density[start_frequency_index,start_direction_index]==min_val:
                assigned = True
                partition_label[start_frequency_index,start_direction_index] = current_label

    if assigned:
        proximate_partitions[current_label] = numpy.array([0], dtype='int64')

    for start_frequency_index in range(0, number_of_frequencies):
        for start_direction_index in range(0, number_of_directions):
            # if already assigned - continue
            if partition_label[
                start_frequency_index, start_direction_index] > NOT_ASSIGNED:
                continue

            ii = [start_frequency_index]
            jj = [start_direction_index]
            proximate_partitions_work = []
            while True:

                direction_index = jj[-1]
                frequency_index = ii[-1]

                neighbour_frequency_indices, neighbour_direction_indices = neighbours(
                    direction_index, frequency_index,
                    number_of_directions, number_of_frequencies)

                node_value = spectral_density[frequency_index, direction_index]

                delta = numpy.zeros_like(neighbour_frequency_indices,dtype='float64')
                for index, ifreq, idir in zip(
                        range(0, len(neighbour_frequency_indices)),
                        neighbour_frequency_indices,
                        neighbour_direction_indices):

                    delta[index] = (spectral_density[ifreq, idir] - node_value)

                    if partition_label[ifreq, idir] > NOT_ASSIGNED:
                        proximate_partitions_work.append(
                            partition_label[ifreq, idir])

                if numpy.all(delta <= 0) or (partition_label[
                                                 frequency_index, direction_index] > NOT_ASSIGNED):

                    # this is a peak, or already leads to a peak
                    if partition_label[
                        frequency_index, direction_index] == NOT_ASSIGNED:
                        current_label += 1
                        label = current_label
                        proximate_partitions[label] = numpy.array(
                            proximate_partitions_work, dtype='int64')
                    else:
                        label = partition_label[
                            frequency_index, direction_index]
                        proximate_partitions[
                            label] = numpy.append(proximate_partitions[
                                                      label],
                                                  proximate_partitions_work)

                    for i, j in zip(ii, jj):
                        partition_label[i, j] = label

                    break
                else:
                    steepest_index = numpy.argmax(delta)
                    ii.append(neighbour_frequency_indices[steepest_index])
                    jj.append(neighbour_direction_indices[steepest_index])
                    continue

        #
    for label in proximate_partitions:
        proximate_partitions[label] = numpy.unique(proximate_partitions[label])
        mask = proximate_partitions[label] == label
        proximate_partitions[label] = proximate_partitions[label][~mask]

    for label_source, item in proximate_partitions.items():
        for label_target in item:
            if label_source == label_target:
                continue

            if label_source not in proximate_partitions[label_target]:
                proximate_partitions[label_target] = numpy.append(
                    proximate_partitions[label_target], label_source)

    return partition_label, proximate_partitions


def find_label_closest_partition(label,
                                 proximity: typing.Dict[int, typing.List[int]],
                                 partitions: typing.Dict[int, WaveSpectrum2D]):
    """
    Find the partition whose peak is closest to the partition under considiration
    Only proximate partitions are considered. Basically: find the partition whose
    maximum is closest to the peak of the current partition.

    :param index:
    :param partitions:
    :return:
    """

    # find the labels of the neighbouring partitions.
    proximate_labels = proximity[label]

    # find the wavenumbers associated with the peak
    kx = partitions[label].peak_wavenumber_east()
    ky = partitions[label].peak_wavenumber_north()

    # Initialize the numpy arrays for distance and indices.
    distance = numpy.zeros((len(proximate_labels),))
    indices = numpy.zeros((len(proximate_labels),), dtype='int32')

    for ii, local_label in enumerate(proximate_labels):
        delta_kx = partitions[local_label].peak_wavenumber_east() - kx
        delta_ky = partitions[local_label].peak_wavenumber_north() - ky
        distance[ii] = numpy.sqrt(delta_kx ** 2 + delta_ky ** 2)
        indices[ii] = local_label

    return indices[numpy.argmin(distance)]


def filter_for_low_energy(partitions: typing.Dict[int, WaveSpectrum2D],
                          proximity: typing.Dict[int, typing.List[int]],
                          total_energy, config):
    for label in list(partitions.keys()):

        if label not in partitions:
            continue

        partition = partitions[label]

        if partition.hm0() == 0:
            closest_label = proximity[label][0]
        else:
            closest_label = find_label_closest_partition(
                label, proximity, partitions)

        if (partition.m0() < config['minimumEnergyFraction'] * total_energy):
            #
            # Merge with the closest partition
            merge_partitions(label, closest_label, partitions,
                             proximity)


def merge_partitions(
        source_label,
        target_label,
        partitions: typing.Dict[int, WaveSpectrum2D],
        proximity: typing.Dict[int, typing.List[int]]) -> WaveSpectrum2D:
    """
    Merge one partition into another partition. (**Has side-effects**)
    :param source_index: source index of the partition to merge in the list
    :param target_index: target index of the partition that is merged into
    :param partitions: the list of partitions
    :param partition_label_array: 2D array of labels indicating to what partition
    the cel belongs

    :return: None
    """

    # Get source/target partitions
    partitions[target_label] = partitions[target_label] + partitions[
        source_label]
    partitions.pop(source_label)

    for proximate_labels_to_source_label in proximity[source_label]:
        proximate_to_update = proximity[proximate_labels_to_source_label]
        index = proximate_to_update.index(source_label)

        if proximate_labels_to_source_label == target_label:
            proximity[target_label].pop(index)
            continue

        #
        if proximate_labels_to_source_label not in proximity[target_label]:
            proximity[target_label].append(proximate_labels_to_source_label)

        if target_label in proximate_to_update:
            proximate_to_update.pop(index)
        else:
            proximate_to_update[index] = target_label

    proximity.pop(source_label)
    return partitions[target_label]


def partition_spectrum(spectrum: WaveSpectrum2D, config=None) -> \
        typing.Tuple[typing.Dict[int, WaveSpectrum2D], typing.Dict[
            int, typing.List[int]]]:
    """
    Create partitioned spectra from a given 2D wavespectrum
    :param spectrum: 2D wavespectrum
    :param config: configuration parameters - see "default_partition_config"
    default variable.

    :return:
    """

    # Update config if provided
    config = default_partition_config | (config if config else {})
    # Make sure there are no NaN (undifined) values
    density = spectrum.variance_density.copy()
    density[numpy.isnan(density)] = 0.0

    # Get partition descriptions using floodfill. This returns an array that
    # labels for each cell in the spectrum to which partition it belongs, an
    # ordered list with the label as index containing which partitions are
    # adjacent to the current partition and the labels.
    partition_label, proximate = floodfill(
        spectrum.frequency, spectrum.direction * numpy.pi / 180, density)

    # We need to convert back to a list, and we need to do it in a somewhat
    # roundabout way as numba does not support dicts of lists (note, proximate
    # currently is a "numba" dictionarly of numpy arrays)
    proximate_partitions = {}
    for label in proximate:
        proximate_partitions[label] = list(proximate[label])

    # Create Partition objects given the label array from the floodfill.
    partitions = {}
    # Create a dict of all draft partitions
    for label, proximity_list in proximate_partitions.items():
        mask = partition_label == label
        partitions[label] = spectrum.extract(mask,
                                             config['minimumDensityRatio'])
    #
    # # merge low energy partitions
    filter_for_low_energy(partitions, proximate_partitions, spectrum.m0(),
                          config)
    #  if there are multiple sea spectra merge them into a single sea-spectrum
    # merge_sea_spectra(partitions, proximate_partitions, config)

    # Return the partitions and the label array.
    return partitions, proximate_partitions


def partition_marker_array(
        partitions: typing.Dict[int, WaveSpectrum2D]) -> numpy.ndarray:
    a_label = list(partitions.keys())[0]
    marker_array = numpy.zeros(
        partitions[a_label].variance_density.shape) + numpy.nan
    for label, partition in partitions.items():
        mask = ~partition.variance_density.mask
        marker_array[mask] = label
    return marker_array


def sum_partitions(
        partitions: typing.Dict[int, WaveSpectrum2D]) -> WaveSpectrum2D:
    key = list(partitions.keys())[0]
    sum_spec = empty_spectrum2D_like(partitions[key])
    for _, spec in partitions.items():
        sum_spec = spec + sum_spec
    return sum_spec


def is_neighbour(partition_1: WaveSpectrum2D,
                 partition_2: WaveSpectrum2D) -> int:
    footprint = generate_binary_structure(partition_2.variance_density.ndim, 2)

    mask_1 = maximum_filter(partition_1.variance_density.filled(0),
                            footprint=footprint,
                            mode='wrap')
    mask_2 = partition_2.variance_density.filled(0)

    return numpy.nansum(mask_1 * mask_2)



from roguewave.wavespectra.spectrum1D import WaveSpectrum1D, \
    WaveSpectrum1DInput
import typing
from typing import Dict,List,Union, overload
from pysofar.spotter import Spotter, SofarApi
from datetime import datetime
from roguewave.tools import datetime_to_iso_time_string, to_datetime
from datetime import timedelta
from .exceptions import ExceptionNoFrequencyData
from pandas import read_csv, to_numeric
import numpy
import os

MAX_LOCAL_LIMIT = 20

def _get_spectrum_from_sofar_spotter_api(
        spotter: Spotter,
        start_date: typing.Union[datetime, str] = None,
        end_date: typing.Union[datetime, str] = None,
        limit: int = MAX_LOCAL_LIMIT,
) -> typing.List[WaveSpectrum1D]:
    """
    Grabs the requested spectra for this spotter based on the given keyword arguments

    :param limit: The limit for data to grab. Defaults to 20, For frequency data max of 100 samples at a time.
    :param start_date: ISO 8601 formatted date string. If not included defaults to beginning of spotters history
    :param end_date: ISO 8601 formatted date string. If not included defaults to end of spotter history

    :return: Data as a FrequencyDataList Object
    """


    start_date = to_datetime(start_date)
    end_date = to_datetime(end_date)

    json_data = spotter.grab_data(
        limit=limit,
        start_date=datetime_to_iso_time_string(start_date),
        end_date=datetime_to_iso_time_string(end_date),
        include_frequency_data=True,
        include_directional_moments=True)

    if not json_data['frequencyData']:
        raise ExceptionNoFrequencyData(
            f'Spotter {spotter.id} has no spectral data for the requested time range')

    out = []
    for spectrum in json_data['frequencyData']:
        # Load the data into the input object- this is a one-to-one
        # mapping of keys between dictionaries (we would not _need_ to
        # map onto a input object-> we could pass spectrum directly to
        # the constructor. But it is clearer to fail here if something
        # is amiss with the api JSON response format)
        wave_spectrum_input = WaveSpectrum1DInput(**spectrum)

        out.append(WaveSpectrum1D(wave_spectrum_input))

    return out

@overload
def get_spectrum_from_sofar_spotter_api(
        spotter_ids: List[str],
        start_date: Union[datetime, str] = None,
        end_date: Union[datetime, str] = None,
        session: SofarApi=None,
        verbose = False,
        limit=None
) -> Dict[str, List[WaveSpectrum1D]]: ...

@overload
def get_spectrum_from_sofar_spotter_api(
        spotter_ids: str,
        start_date: Union[datetime, str] = None,
        end_date: Union[datetime, str] = None,
        session: SofarApi=None,
        verbose = False,
        limit=None
) -> List[WaveSpectrum1D]: ...

def get_spectrum_from_sofar_spotter_api(
        spotter_ids: Union[str,List[str]],
        start_date: Union[datetime, str] = None,
        end_date: Union[datetime, str] = None,
        session: SofarApi=None,
        verbose = False,
        limit=None
) -> Union[ Dict[str, List[WaveSpectrum1D]], List[WaveSpectrum1D]]:
    """
    Grabs the requested spectra for this spotter based on the given keyword arguments

    :param limit: The limit for data to grab. Defaults to 20, For frequency data max of 100 samples at a time.
    :param start_date: ISO 8601 formatted date string. If not included defaults to beginning of spotters history
    :param end_date: ISO 8601 formatted date string. If not included defaults to end of spotter history

    :return: Data as a FrequencyDataList Object
    """

    if not isinstance( spotter_ids, list ):
        spotter_ids = [spotter_ids]

    if session is None:
        session = SofarApi()


    data = {}
    for spotter_id in spotter_ids:
        spotter = Spotter(spotter_id,spotter_id,session=session)
        _start_date = start_date
        data[spotter_id] = []
        while True:
            if limit:
                if len(data) >= limit:
                    break

                local_limit = min( limit-len(data), MAX_LOCAL_LIMIT)
            else:
                local_limit = MAX_LOCAL_LIMIT

            try:
                next = _get_spectrum_from_sofar_spotter_api(spotter,_start_date,end_date,local_limit)
            except ExceptionNoFrequencyData as e:
                if not len(data[spotter_id]):
                    raise e
                else:
                    break

            data[spotter_id] += next
            if len(next) < local_limit:
                break
            else:
                _start_date = to_datetime(next[-1].timestamp) + timedelta(seconds=900)

    if len(spotter_ids) == 1:
        return data[spotter_ids[0]]
    else:
        return data


def get_spectrum_from_parser_output(path: str)->typing.List[WaveSpectrum1D]:
    """

    :param path: Path that contains the output from the spotter parser.
    :return: A list of WaveSpectrum1D objects.
    """
    def load_spectral_file(file):
        if os.path.isfile(file):
            data = read_csv(file).apply(to_numeric, errors='coerce')
        else:
            raise FileNotFoundError(file)

        columns = list(data.columns)
        frequencies = [float(x) for x in columns[8:]]
        values = data[columns[8:]].values

        time_tuple = data[columns[0:6]].values
        time = []
        for index in range(time_tuple.shape[0]):
            time.append(
                datetime(year=time_tuple[index, 0], month=time_tuple[index, 1],
                         day=time_tuple[index, 2], hour=time_tuple[index, 3],
                         minute=time_tuple[index, 4],
                         second=time_tuple[index, 5]))

        return {'time': numpy.array(time), 'frequencies': frequencies,
                'values': values}

    files = ['a1', 'b1', 'a2', 'b2', 'Szz']
    data = {}
    for file_type in files:
        file_location = os.path.join(path, file_type+'.csv')
        data[file_type] = load_spectral_file(file_location)

    number_of_spectra = data['Szz']['values'].shape[0]
    spectra = []
    for index in range(0,number_of_spectra):
        input = WaveSpectrum1DInput(frequency=data['Szz']['frequencies'],
                                    varianceDensity=data['Szz']['values'][index,:],
                                    timestamp=data['Szz']['time'][index], latitude=None,
                                    longitude=None, a1=data['a1']['values'][index,:],
                                    b1=data['b1']['values'][index,:],
                                    a2=data['a2']['values'][index,:],
                                    b2=data['b2']['values'][index,:])
        spectra.append(WaveSpectrum1D(input))
    return spectra
